# Date conversion utilities for sanmiao

import re
import pandas as pd
from math import floor
from typing import Tuple, Union

from .config import normalize_defaults, get_phrase_dic
from .loaders import load_csv, load_num_tables
from .utils import guess_variant

# Ganzhi conversion constants
_GANZHI_ZH_TO_NUM = {
    '甲子': 1, '乙丑': 2, '丙寅': 3, '丁卯': 4, '戊辰': 5, '己巳': 6, '庚午': 7, '辛未': 8, '壬申': 9, '癸酉': 10,
    '甲戌': 11, '乙亥': 12, '丙子': 13, '丁丑': 14, '戊寅': 15, '己卯': 16, '庚辰': 17, '辛巳': 18, '壬午': 19, '癸未': 20,
    '甲申': 21, '乙酉': 22, '丙戌': 23, '丁亥': 24, '戊子': 25, '己丑': 26, '庚寅': 27, '辛卯': 28, '壬辰': 29, '癸巳': 30,
    '甲午': 31, '乙未': 32, '丙申': 33, '丁酉': 34, '戊戌': 35, '己亥': 36, '庚子': 37, '辛丑': 38, '壬寅': 39, '癸卯': 40,
    '甲辰': 41, '乙巳': 42, '丙午': 43, '丁未': 44, '戊申': 45, '己酉': 46, '庚戌': 47, '辛亥': 48, '壬子': 49, '癸丑': 50,
    '甲寅': 51, '乙卯': 52, '丙辰': 53, '丁巳': 54, '戊午': 55, '己未': 56, '庚申': 57, '辛酉': 58, '壬戌': 59, '癸亥': 60,
}
_NUM_TO_GANZHI_ZH = {v: k for k, v in _GANZHI_ZH_TO_NUM.items()}

_GANZHI_PINYIN_TO_NUM = {
    'jiazi': 1, 'yichou': 2, 'bingyin': 3, 'dingmao': 4, 'wuchen': 5, 'jisi': 6, 'gengwu': 7, 'xinwei': 8, 'renshen': 9, 'guiyou': 10,
    'jiaxu': 11, 'yihai': 12, 'bingzi': 13, 'dingchou': 14, 'wuyin': 15, 'jimao': 16, 'gengchen': 17, 'xinsi': 18, 'renwu': 19, 'guiwei': 20,
    'jiashen': 21, 'yiyou': 22, 'bingxu': 23, 'dinghai': 24, 'wuzi': 25, 'jichou': 26, 'gengyin': 27, 'xinmao': 28, 'renchen': 29, 'guisi': 30,
    'jiawu': 31, 'yiwei': 32, 'bingshen': 33, 'dingyou': 34, 'wuxu': 35, 'jihai': 36, 'gengzi': 37, 'xinchou': 38, 'renyin': 39, 'guimao': 40,
    'jiachen': 41, 'yisi': 42, 'bingwu': 43, 'dingwei': 44, 'wushen': 45, 'jiyou': 46, 'gengxu': 47, 'xinhai': 48, 'renzi': 49, 'guichou': 50,
    'jiayin': 51, 'yimao': 52, 'bingchen': 53, 'dingsi': 54, 'wuwu': 55, 'jiwei': 56, 'gengshen': 57, 'xinyou': 58, 'renxu': 59, 'guihai': 60,
}
_NUM_TO_GANZHI_PINYIN = {v: k for k, v in _GANZHI_PINYIN_TO_NUM.items()}


def gz_year(num: int) -> int:
    """
    Converts Western calendar year to sexagenary year (numerical)
    :param num: int
    :return: int
    """
    x = (num - 4) % 60 + 1
    return x


def jdn_to_gz(jdn: int, en: bool = False) -> str:
    """
    Convert from Julian day number (JDN) to sexagenary day, with output in Pinyin (en=True) or Chinese (en=False).
    :param jdn: float
    :param en: bool
    """
    jdn = int(jdn - 9.5) % 60
    if jdn == 0:
        jdn = 60
    gz = ganshu(jdn, en)
    return gz


def ganshu(gz_in, en=False):
    """
    Convert from sexagenary counter (string) to number (int) and vice versa.
    :param gz_in: str, int, or float
    :param en: Boolean, whether into Pinyin (vs Chinese)
    :return: int or str
    """

    if en:
        to_num = _GANZHI_PINYIN_TO_NUM
        to_str = _NUM_TO_GANZHI_PINYIN
    else:
        to_num = _GANZHI_ZH_TO_NUM
        to_str = _NUM_TO_GANZHI_ZH

    # string -> number
    if isinstance(gz_in, str):
        s = gz_in.strip()
        # Normalise for Chen dynasty taboo
        if not en:
            s = re.sub('景', '丙', s)
            return to_num.get(s, None)
        else:
            s = s.lower()
        out = to_num.get(s, None)
        if out is None:
            raise ValueError(f"Invalid ganzhi string: {gz_in}")
        return out

    # number -> string
    try:
        n = int(gz_in)
    except (TypeError, ValueError):
        raise ValueError(f"Invalid ganzhi string: {gz_in}")
        return None

    return to_str.get(n, None)


def numcon(x):
    """
    Convert Chinese numerals into arabic numerals (from 9999 down) and from arabic into Chinese (from 99 down)
    :param x: str, int, or float
    :return: int
    """
    chinese_numerals = '〇一二三四五六七八九'
    if isinstance(x, str):  # If string
        if x in ['正月', '元年']:
            return 1
        else:
            # Normalize number string
            tups = [
                ('元', '一'),
                ('廿', '二十'), ('卅', '三十'), ('卌', '四十'), ('兩', '二'),
                ('初', '〇'), ('無', '〇'), ('卄', '二十'), ('丗', '三十')
            ]
            for tup in tups:
                x = re.sub(tup[0], tup[1], x)
            # Variables
            arab_numerals = '0123456789'
            w_place_values = ['一', '二', '三', '四', '五', '六', '七', '八', '九', '十', '〇', '百', '千', '萬']
            # Remove all non number characters
            only_numbers = ''
            for char in x:
                if char in w_place_values:
                    only_numbers += char
            # Convert to Frankenstein string
            frankenstein = only_numbers.translate(str.maketrans(chinese_numerals, arab_numerals))
            # Determine if place value words occur
            place_values = ['十', '百', '千', '萬']
            count = 0
            for i in place_values:
                if i in frankenstein:
                    count = 1
                    break
            # Logic tree
            if count == 0:  # If there are no place values
                # Try to return as integer
                if frankenstein.strip():  # Only try to convert non-empty strings
                    try:
                        return int(frankenstein)
                    except (ValueError, TypeError):
                        return None
                else:
                    return None
            else:  # If there are place value words
                # Remove zeros
                frankenstein = frankenstein.replace('0', '')
                # Empty result to which to add each place value
                numeral = 0
                # Thousands
                thousands = frankenstein.split('千')
                if len(thousands) == 2 and len(thousands[0]) == 0:
                    numeral += 1000
                elif len(thousands) == 2 and len(thousands[0]) == 1:
                    numeral += 1000 * int(thousands[0])
                # Hundreds
                hundreds = thousands[-1].split('百')
                if len(hundreds) == 2 and len(hundreds[0]) == 0:
                    numeral += 100
                elif len(hundreds) == 2 and len(hundreds[0]) == 1:
                    numeral += 100 * int(hundreds[0])
                # Tens
                tens = hundreds[-1].split('十')
                if len(tens) == 2 and len(tens[0]) == 0:
                    numeral += 10
                elif len(tens) == 2 and len(tens[0]) == 1:
                    numeral += 10 * int(tens[0])
                remainder = tens[-1]
                # Units
                try:
                    numeral += int(remainder[0])
                except (IndexError, ValueError):
                    # If remainder is empty or not a digit, skip it
                    pass
                return int(numeral)
    else:  # To convert from integer/float to Chinese
        x = int(x)
        # Blank string
        s = ''
        # Find number of thousands
        x %= 10000
        thousands = x // 1000
        if thousands > 0:
            if thousands > 1:
                s += chinese_numerals[thousands]
            s += '千'
        # Find number of hundreds
        x %= 1000
        hundreds = x // 100
        if hundreds > 0:
            if hundreds > 1:
                s += chinese_numerals[hundreds]
            s += '百'
        # Find number of tens
        x %= 100
        tens = x // 10
        if tens > 0:
            if tens > 1:
                s += chinese_numerals[tens]
            s += '十'
        # Find units
        rem = int(x % 10)
        if rem > 0:
            s += chinese_numerals[rem]
        return s


def iso_to_jdn(date_string, proleptic_gregorian=False, gregorian_start=None):
    """
    Convert a date string (YYYY-MM-DD) to a Julian Day Number (JDN).

    :param date_string: str (date in "YYYY-MM-DD" format, e.g., "2023-01-01" or "-0044-03-15")
    :param proleptic_gregorian: bool
    :param gregorian_start: list
    :return: float (Julian Day Number) or None if invalid
    """
    # Defaults
    gregorian_start, civ = normalize_defaults(gregorian_start)

    # Validate inputs
    if not re.match(r'^-?\d+-\d+-\d+$', date_string):
        return None

    try:
        # Handle negative year
        if date_string[0] == '-':
            mult = -1
            date_string = date_string[1:]
        else:
            mult = 1

        # Split and convert to integers
        year, month, day = map(int, date_string.split("-"))
        year *= mult

        # Validate month and day
        if not (1 <= month <= 12) or not (1 <= day <= 31):  # Basic validation
            return None

        # Determine calendar for historical mode
        is_julian = False
        a, b, c = gregorian_start
        if not proleptic_gregorian:
            if year < a:
                is_julian = True
            elif year == a and month < b:
                is_julian = True
            elif year == a and month == b and day <= c:
                is_julian = True

        # Adjust months and years so March is the first month
        if month <= 2:
            year -= 1
            month += 12

        # Calculate JDN
        if proleptic_gregorian or not is_julian:
            # Gregorian calendar
            a = floor(year / 100)
            b = floor(a / 4)
            c = 2 - a + b
            jdn = floor(365.25 * (year + 4716)) + floor(30.6001 * (month + 1)) + day + c - 1524.5
        else:
            # Julian calendar
            jdn = floor(365.25 * (year + 4716)) + floor(30.6001 * (month + 1)) + day - 1524.5

        return jdn
    except ValueError:
        return None


def jdn_to_iso(jdn, proleptic_gregorian=False, gregorian_start=None):
    """
    Convert a Julian Day Number (JDN) to a date string (YYYY-MM-DD).

    :param jdn: int or float (e.g., 2299159.5 = 1582-10-15)
    :param proleptic_gregorian: bool
    :param gregorian_start: list
    :return: str (ISO date string) or None if invalid
    """
    # Defaults
    gregorian_start, civ = normalize_defaults(gregorian_start)

    # Get Gregorian reform JDN
    gs_str = f"{gregorian_start[0]}-{gregorian_start[1]}-{gregorian_start[2]}"
    gs_jdn = iso_to_jdn(gs_str, proleptic_gregorian, gregorian_start)
    if not isinstance(jdn, (int, float)):
        return None
    try:
        jdn = floor(jdn + 0.5)
        is_julian = not proleptic_gregorian and jdn < gs_jdn
        if proleptic_gregorian or not is_julian:
            a = jdn + 32044
            b = floor((4 * a + 3) / 146097)
            c = a - floor((146097 * b) / 4)
            d = floor((4 * c + 3) / 1461)
            e = c - floor((1461 * d) / 4)
            m = floor((5 * e + 2) / 153)
            day = e - floor((153 * m + 2) / 5) + 1
            month = m + 3 - 12 * floor(m / 10)
            year = 100 * b + d - 4800 + floor(m / 10)
        else:
            a = jdn + 32082
            b = floor((4 * a + 3) / 1461)
            c = a - floor((1461 * b) / 4)
            m = floor((5 * c + 2) / 153)
            day = c - floor((153 * m + 2) / 5) + 1
            month = m + 3 - 12 * floor(m / 10)
            year = b - 4800 + floor(m / 10)
        if year <= 0:
            year_str = f"-{abs(year):04d}"
        else:
            year_str = f"{year:04d}"
        date_str = f"{year_str}-{month:02d}-{day:02d}"
        if not re.match(r'^-?\d{4}-\d{2}-\d{2}$', date_str):
            return None
        return date_str
    except (ValueError, OverflowError):
        return None


def jdn_to_ccs(x, by_era=True, proleptic_gregorian=False, gregorian_start=None, lang='en', civ=None):
    """
    Convert Julian Day Number to Chinese calendar string.
    :param x: float (Julian Day Number) or str (ISO date string Y-M-D)
    :param by_era: bool (filter from era JDN vs index year)
    :param proleptic_gregorian: bool
    :param gregorian_start: list
    :param lang: str, language ('en' or 'fr')
    :param civ: str ('c', 'j', 'k') or list (['c', 'j', 'k']) to filter by civilization
    :return output_string: str
    """
    # Defaults
    gregorian_start, civ = normalize_defaults(gregorian_start, civ)
    if lang is None:
        lang = 'en'
    phrase_dic = get_phrase_dic(lang)
    if isinstance(x, str):
        iso = x
        jdn = iso_to_jdn(x, proleptic_gregorian, gregorian_start)
    else:
        jdn = x
        iso = jdn_to_iso(jdn, proleptic_gregorian, gregorian_start)
    output_string = f'{phrase_dic.get("ui")}: {iso} (JD {jdn})\n{phrase_dic.get("matches")}:\n'
    # Load CSV tables
    era_df, dyn_df, ruler_df, lunar_table = load_num_tables(civ=civ)
    ruler_tag_df = load_csv('rul_can_name.csv')[['person_id', 'string']]
    # Filter ruler_tag_df by filtered rulers
    if not ruler_df.empty:
        valid_person_ids = ruler_df['person_id'].unique()
        ruler_tag_df = ruler_tag_df[ruler_tag_df['person_id'].isin(valid_person_ids)]
    # Filter lunar table by JDN
    lunar_table = lunar_table[(lunar_table['nmd_jdn'] <= jdn) & (lunar_table['hui_jdn'] + 1 > jdn)]
    #
    if by_era:
        # Filter era dataframe by JDN
        df = era_df[(era_df['era_start_jdn'] <= jdn) & (era_df['era_end_jdn'] > jdn)].drop_duplicates(subset=['era_id'])
        df = df[['dyn_id', 'cal_stream', 'era_id', 'ruler_id', 'era_name', 'era_start_year']].rename(columns={'ruler_id': 'person_id'})
        # Get ruler names
        df = df.merge(ruler_tag_df, how='left', on='person_id')
        df = df.rename(columns={'person_id': 'ruler_id', 'string': 'ruler_name'})
        # Get dynasty names
        df = df.merge(dyn_df[['dyn_id', 'dyn_name']], how='left', on='dyn_id')
        # Merge with lunar table
        lunar_table = df.merge(lunar_table, how='left', on='cal_stream')
        # Add ruler start year, just to be safe
        temp = ruler_df[['person_id', 'emp_start_year']]
        temp = temp.rename(columns={'person_id': 'ruler_id'})
        lunar_table = lunar_table.merge(temp, how='left', on='ruler_id')
    else:
        # Merge dynasties
        lunar_table = lunar_table.merge(dyn_df, how='left', on='cal_stream')
        # Filter by index year
        lunar_table = lunar_table[lunar_table['dyn_start_year'] <= lunar_table['ind_year']]
        lunar_table = lunar_table[lunar_table['dyn_end_year'] > lunar_table['ind_year']]
        del lunar_table['dyn_start_year'], lunar_table['dyn_end_year']
        # Merge rulers
        del ruler_df['cal_stream'], ruler_df['max_year']
        lunar_table = lunar_table.merge(ruler_df, how='left', on='dyn_id')
        # Merge ruler tags
        lunar_table = lunar_table.merge(ruler_tag_df, how='left', on='person_id')
        lunar_table = lunar_table.rename(columns={'person_id': 'ruler_id', 'string': 'ruler_name'})
        # Filter by index year
        lunar_table = lunar_table[lunar_table['emp_start_year'] <= lunar_table['ind_year']]
        lunar_table = lunar_table[lunar_table['emp_end_year'] > lunar_table['ind_year']]
        del lunar_table['emp_end_year']
        # Clean eras
        del era_df['max_year']
        era_df = era_df.drop_duplicates(subset=['era_id'])
        # Merge eras
        lunar_table = lunar_table.merge(era_df, how='left', on=['dyn_id', 'cal_stream', 'ruler_id'])
        # Filter by index year
        lunar_table = lunar_table[lunar_table['era_start_year'] <= lunar_table['ind_year']]
        lunar_table = lunar_table[lunar_table['era_end_year'] > lunar_table['ind_year']]
        del lunar_table['era_end_year']
    if not lunar_table.empty:
        lunar_table = lunar_table.sort_values(by=['cal_stream', 'dyn_id'])        
        """
        Note: where era and ruler start years differ, I sometimes get duplicates:

        ,dyn_id,cal_stream,era_id,ruler_id,era_name,era_start_year,ruler_name,dyn_name,ind_year,year_gz,month,intercalary,nmd_gz,nmd_jdn,hui_jdn,max_day,hui_gz,emp_start_year
        0,124,3.0,636,15353.0,至正,1341,順帝妥懽帖睦爾,元,1342,19,1,0,10,2211259.5,2211287.5,29.0,辛丑,1333.0
        1,133,4.0,930,16394.0,興国,1340,後村上天皇,日本,1342,19,1,0,10,2211259.5,2211288.5,30.0,壬寅,1339.0
        2,133,4.0,939,16398.0,暦応,1338,光明天皇,日本,1342,19,1,0,10,2211259.5,2211288.5,30.0,壬寅,1336.0
        3,141,8.0,1175,16597.0,後元,1340,忠惠王,高麗,1342,19,1,0,10,2211259.5,2211287.5,29.0,辛丑,1331.0
        4,141,8.0,1175,16597.0,後元,1340,忠惠王,高麗,1342,19,1,0,10,2211259.5,2211287.5,29.0,辛丑,1340.0

        This merits rethinking, but the following will work for now.
        """
        lunar_table = lunar_table.drop_duplicates(subset=['era_id'])
        # Create strings
        for index, row in lunar_table.iterrows():
            # Output dynasty and ruler name
            output_string += f"{row['dyn_name']}{row['ruler_name']}"
            # Find Julian year
            iso_string = jdn_to_iso(jdn, proleptic_gregorian, gregorian_start)
            if iso_string[0] == '-':
                iso_string = iso_string[1:]
                mult = -1
            else:
                mult = 1
            year = int(re.split('-', iso_string)[0]) * mult
            # Convert to era or ruler year
            # Check if era_start_year is valid (not NaN) - works for both int and float
            if pd.notna(row['era_start_year']):
                # We have a valid era, use it (even if era_name is blank)
                if isinstance(row['era_name'], str) and row['era_name'] != '':
                    output_string += f"{row['era_name']}"
                # Find era year
                era_year = year - int(row['era_start_year']) + 1
                era_year = numcon(era_year) + '年'
                if era_year == "一年":
                    era_year = "元年"
                output_string += era_year
            else:
                # No valid era, fall back to ruler start year
                ruler_year = year - row['emp_start_year'] + 1
                ruler_year = numcon(ruler_year) + '年'
                if ruler_year == "一年":
                    ruler_year = "元年"
                output_string += ruler_year
            # Sexegesimal year
            sex_year = ganshu(row['year_gz'])
            output_string += f"（歲在{sex_year}）"
            # Month
            if row['intercalary'] == 1:
                output_string += '閏'
            if row['month'] == 1:
                month = '正月'
            elif row['month'] == 13:
                month = '臘月'
            elif row['month'] == 14:
                month = '一月'
            else:
                month = numcon(row['month']) + '月'
            output_string += month
            # Find day
            if int(jdn - .5) + .5 == row['nmd_jdn']:
                day = '朔'
            elif int(jdn - .5) + .5 == row['hui_jdn']:
                num = numcon(row['hui_jdn'] - row['nmd_jdn'] + 1) + '日'
                day = f"晦（{num}）"
            else:
                day = numcon(int(jdn - row['nmd_jdn']) + 1) + '日'
            output_string += day
            # Sexagenary day
            output_string += jdn_to_gz(jdn)
            # Line break
            output_string += '\n'
        output_string = output_string[:-1]
        # Output
        return output_string
    else:
        return None


def jy_to_ccs(y, lang='en', civ=None):
    """
    Convert Western year to Chinese calendar string.
    :param y: int
    :param lang: str, language ('en', 'fr', 'zh', 'ja', 'de'). Defaults to 'en' if not specified or invalid.
    :param civ: str ('c', 'j', 'k') or list (['c', 'j', 'k']) to filter by civilization
    :return output_string: str
    """
    _gs, civ = normalize_defaults(None, civ)
    if lang is None:
        lang = 'en'
    phrase_dic = get_phrase_dic(lang)
    
    # Year format strings for different languages
    if y > 0:
        if lang == 'en':
            fill = f"A.D. {int(y)}"
        elif lang == 'fr':
            fill = f"{int(y)} apr. J.-C."
        elif lang == 'de':
            fill = f"{int(y)} n. Chr."
        elif lang == 'zh':
            fill = f"公元{int(y)}年"
        elif lang == 'ja':
            fill = f"西暦{int(y)}年"
        else:
            fill = f"A.D. {int(y)}"  # Default to English
    else:
        if lang == 'en':
            fill = f"{int(abs(y)) + 1} B.C."
        elif lang == 'fr':
            fill = f"{int(abs(y)) + 1} av. J.-C."
        elif lang == 'de':
            fill = f"{int(abs(y)) + 1} v. Chr."
        elif lang == 'zh':
            fill = f"公元前{int(abs(y)) + 1}年"
        elif lang == 'ja':
            fill = f"紀元前{int(abs(y)) + 1}年"
        else:
            fill = f"{int(abs(y)) + 1} B.C."  # Default to English
    output_string = f'{phrase_dic.get("ui")}: {y} ({fill})\n{phrase_dic.get("matches")}:\n'
    # Load CSV tables
    era_df, dyn_df, ruler_df, lunar_table = load_num_tables(civ=civ)
    ruler_tag_df = load_csv('rul_can_name.csv')[['person_id', 'string']]
    # Filter ruler_tag_df by filtered rulers
    if not ruler_df.empty:
        valid_person_ids = ruler_df['person_id'].unique()
        ruler_tag_df = ruler_tag_df[ruler_tag_df['person_id'].isin(valid_person_ids)]
    ruler_tag_df = ruler_tag_df[['person_id', 'string']]
    # Filter dynasties by year
    df = dyn_df[(dyn_df['dyn_start_year'] <= y) & (dyn_df['dyn_end_year'] >= y)]
    cols = ['dyn_id', 'dyn_name', 'cal_stream']
    df = df[cols]
    # Merge rulers
    del ruler_df['cal_stream']
    df = df.merge(ruler_df, how='left', on=['dyn_id'])
    # Filter by year
    df = df[(df['emp_start_year'] <= y) & (df['emp_end_year'] >= y)]
    # Merge ruler strings
    df = df.merge(ruler_tag_df, how='left', on='person_id')
    df = df.rename(columns={'person_id': 'ruler_id', 'string': 'ruler_name'})
    cols = ['dyn_id', 'dyn_name', 'cal_stream', 'ruler_id', 'emp_start_year', 'ruler_name']
    df = df[cols]
    # Merge era
    era_df = era_df[['era_id', 'ruler_id', 'era_name', 'era_start_year', 'era_end_year']]
    df = df.merge(era_df, how='left', on='ruler_id')
    # Filter by year
    df = df[(df['era_start_year'] <= y) & (df['era_end_year'] >= y)].sort_values(by=['cal_stream', 'dyn_id'])
    # Filter duplicates
    try:
        df['variant_rank'] = df['era_name'].apply(guess_variant)
        df = (
            df.sort_values(by='variant_rank')
            .drop_duplicates(subset=['ruler_id', 'era_id'], keep="first")
            .drop(columns="variant_rank")
        )
    except TypeError:
        df = df.drop_duplicates(subset=['ruler_id', 'era_id'], keep="first")
    if not df.empty:
        # Create strings
        for index, row in df.iterrows():
            # Output dynasty and ruler name
            output_string += f"{row['dyn_name']}{row['ruler_name']}"
            # Convert to era or ruler year
            if isinstance(row['era_name'], str):
                output_string += f"{row['era_name']}"
                # Find era year
                era_year = y - row['era_start_year'] + 1
                era_year = numcon(era_year) + '年'
                if era_year == "一年":
                    era_year = "元年"
                output_string += era_year
            else:
                ruler_year = y - row['emp_start_year'] + 1
                ruler_year = numcon(ruler_year) + '年'
                if ruler_year == "一年":
                    ruler_year = "元年"
                output_string += ruler_year
            # Sexegesimal year
            sex_year = ganshu(gz_year(y))
            output_string += f"（歲在{sex_year}）"
            # Line break
            output_string += '\n'
        output_string = output_string[:-1]
        # Output
        return output_string
    else:
        return None



import re
import lxml.etree as et
from .loaders import (
    load_csv, load_tag_tables
)
from .config import get_cal_streams_from_civ
from .xml_utils import (
    strip_ws_in_text_nodes, clean_attributes, replace_in_text_and_tail
)

SKIP = {"date","year","month","day","gz","sexYear","era","ruler","dyn","suffix","int","lp",
        "nmdgz","lp_filler","filler","season","gy","rel","meta","pb","text","body"}  # adjust tags you want to skip

SKIP_ALL = {"date","year","month","day","gz","sexYear","era","ruler","dyn","suffix",
            "int","lp","nmdgz","lp_filler","filler","season","gy","rel"}

SKIP_TEXT_ONLY = {"pb", "meta"}

YEAR_RE   = re.compile(r"((?<![一二三四五六七八九])(?:[一二三四五六七八九十]+|十有[一二三四五六七八九]|元)[年載])")
# "廿<date><year>" fix disappears because we won't create that broken boundary in text mode.

# Months: order matters (more specific first)
LEAPMONTH_RE1 = re.compile(r"閏月")
LEAPMONTH_RE2 = re.compile(r"閏((?:十[一二]|十有[一二]|正)月)")
LEAPMONTH_RE3 = re.compile(r"閏((?:[一二三四五六七八九十]|正|臘)月)")
MONTH_RE1     = re.compile(r"((?:十有[一二]|十[一二]|正)月)")
MONTH_RE2     = re.compile(r"((?<![一二三四五六七八九])(?:[一二三四五六七八九十]|正|臘)月)")

DAY_RE    = re.compile(r"((?<![一二三四五六七八九])(?:[一二三四五六七八九]|[一二]*十[一二三四五六七八九]*|[廿卄][一二三四五六七八九]*|卅|丗|三十)日)")
GZ_RE     = re.compile(r"([甲乙丙景丁戊己庚辛壬癸][子丑寅卯辰巳午未申酉戌亥])")
SEXYEAR_RE = re.compile(r"(([甲乙丙景丁戊己庚辛壬癸][子丑寅卯辰巳午未申酉戌亥]))(年|歲)")
SEASON_RE = re.compile(r"([春秋冬夏])")

LP_RE = re.compile(r"([朔晦])")
GY_RE = re.compile(r"(改元)")

# Relational prefixes.
#
# User rules (2026-01):
# - Only "其" and "先是" may be tagged with unit="" (handled structurally later, not by regex).
# - All other rel markers require an explicit unit (年/歲/月); without a unit they MUST NOT be tagged.
# - "是歲，" / "今年，" / "今月，" may precede anything; whether it is *attached into* the following date is decided later.
# - "明月" means "bright moon", not "next month", so we exclude "明" when followed by just "月".
#
# Groups:
#   1 = dir char (後/次/來/明/昨/前/去/其/是/今)
#   2 = unit (年/歲/月), must be present in this regex
#   3 = trailing Chinese comma(s) immediately following the unit (optional)
# Note: Two patterns - one for "明" (must start with 年 or 歲), one for others
REL_RE_MING = re.compile(r"(明)([年歲][月]?)(，*)")  # "明" must be followed by 年 or 歲 (optionally 月 after)
REL_RE_OTHER = re.compile(r"([後次來昨前去其是今])([年歲月]+)(，*)")  # Other direction chars with any unit
REL_RE_XIANSHI = re.compile(r"(先是)(，*)")  # "先是" (previously/before this) - special compound pattern
SEX_YEAR_PREFIX_RE = re.compile(r"(歲[次在])\s*$")
PUNCT_RE = re.compile(r"^[，,、\s]*")

ERA_SUFFIX_RE = re.compile(r"^(之?初|中|之?末|之?季|末年|之?時|之世)")
DYNASTY_SUFFIX_RE = re.compile(r"^(之?初|中|之?末|之?季|末年|之?時|之世)")
RULER_SUFFIX_RE = re.compile(r"^(之?初|中|之?末|之?季|末年|之?時|之世|即位|踐阼)")


def replace_in_text_and_tail(
    xml_root,
    pattern: re.Pattern,
    make_element,
    skip_text_tags=frozenset(),
    skip_all_tags=frozenset(),
):
    """
    Replace pattern matches in text and tail attributes of XML elements.
    Uses iterative approach to handle newly inserted elements properly.
    
    Key point: Even if an element's tag is in skip_all_tags (like <date>),
    we still need to process its TAIL, because that tail might contain
    more patterns that need to be matched.
    """
    # Process elements depth-first, but need to re-scan for new elements
    # Keep processing until no more matches are found
    max_passes = 50  # Safety limit to prevent infinite loops
    changed = True
    
    for pass_num in range(max_passes):
        if not changed:
            break
        changed = False
        
        # Collect all elements to process in this pass
        # Use list() to create snapshot, but we'll re-scan if changes occur
        elements_to_check = []
        for el in xml_root.iter():
            # Always include elements to process their tail
            # We'll skip processing their text/children if tag is in skip_all_tags
            elements_to_check.append(el)
        
        for el in elements_to_check:
            # Skip if element was removed
            parent = el.getparent()
            if parent is None and el is not xml_root:
                continue
            
            # Decide which slots to process
            # CRITICAL: Even if element is in skip_all_tags, we still process its tail!
            # The tail of a <date> element might contain more patterns.
            if el.tag in skip_all_tags:
                # Skip processing text (children) of these elements, but process tail
                slots = ("tail",)
            elif el.tag in skip_text_tags:
                slots = ("tail",)
            else:
                slots = ("text", "tail")

            for slot in slots:
                s = getattr(el, slot)
                if not s or not pattern.search(s):
                    continue

                matches = list(pattern.finditer(s))
                if not matches:
                    continue

                chunks = []
                last = 0
                for m in matches:
                    chunks.append(s[last:m.start()])
                    chunks.append(m)
                    last = m.end()
                chunks.append(s[last:])

                if slot == "text":
                    el.text = chunks[0]
                    pos = 0
                    for i in range(1, len(chunks), 2):
                        new_el = make_element(chunks[i])
                        new_el.tail = chunks[i + 1]
                        el.insert(pos, new_el)
                        pos += 1
                    changed = True
                else:  # tail
                    parent = el.getparent()
                    if parent is None:
                        continue
                    idx = parent.index(el)
                    el.tail = chunks[0]
                    pos = idx + 1
                    for i in range(1, len(chunks), 2):
                        new_el = make_element(chunks[i])
                        new_el.tail = chunks[i + 1]
                        parent.insert(pos, new_el)
                        pos += 1
                    changed = True


def make_simple_date(tagname, group=1):
    """
    Create a function that generates XML date elements with specified tag.

    :param tagname: str, XML tag name for the date element
    :param group: int, regex group number to extract text from
    :return: function that creates XML date elements
    """
    def _mk(m):
        d = et.Element("date")
        c = et.SubElement(d, tagname)
        c.text = m.group(group)
        return d
    return _mk


def make_sexyear(m):
    """
    Create a date element with sexYear structure: <date><sexYear>甲子<filler>年</filler></sexYear></date>
    """
    d = et.Element("date")
    sy = et.SubElement(d, "sexYear")
    sy.text = m.group(1)  # sexagenary part (甲子 etc.)
    filler = et.SubElement(sy, "filler")
    filler.text = m.group(3)  # suffix (年 or 歲)
    return d


def make_leap_month_exact_monthtext(month_text: str):
    """
    Create XML element for leap month with specific month text.

    :param month_text: str, text for the month element
    :return: et.Element, XML date element for leap month
    """
    d = et.Element("date")
    i = et.SubElement(d, "int"); i.text = "閏"
    m = et.SubElement(d, "month"); m.text = month_text
    return d


def make_leapmonth_from_group1(m):
    """
    Create leap month element from regex match group 1.

    :param m: regex match object
    :return: et.Element, XML date element for leap month
    """
    return make_leap_month_exact_monthtext(m.group(1))


def make_leapmonth_yue(m):
    # "閏月" -> <date><int>閏</int><month>月</month></date>
    return make_leap_month_exact_monthtext("月")


def tag_basic_tokens(xml_root):
    """
    Tag basic date tokens (year, month, day, etc.) in XML tree.

    :param xml_root: et.Element, root of XML tree to process
    :return: et.Element, modified XML root with tagged date elements
    """
    # year
    replace_in_text_and_tail(xml_root, YEAR_RE, make_simple_date("year"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # leap month variants (specific -> general)
    replace_in_text_and_tail(xml_root, LEAPMONTH_RE1, make_leapmonth_yue, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, LEAPMONTH_RE2, make_leapmonth_from_group1, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, LEAPMONTH_RE3, make_leapmonth_from_group1, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # month (specific -> general)
    replace_in_text_and_tail(xml_root, MONTH_RE1, make_simple_date("month"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, MONTH_RE2, make_simple_date("month"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # sexagenary year (before gz to avoid conflicts)
    replace_in_text_and_tail(xml_root, SEXYEAR_RE, make_sexyear, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # day, gz, season
    replace_in_text_and_tail(xml_root, DAY_RE, make_simple_date("day"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, GZ_RE, make_simple_date("gz"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, SEASON_RE, make_simple_date("season"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    return xml_root


def promote_gz_to_sexyear(xml_root):
    """
    Promote sexagenary day (gz) elements to sexagenary year (sexYear) when:
    1. Preceded by explicit year markers like 歲次 or 歲在, OR
    2. Followed by 年 or 歲 (e.g., "甲子年" where gz wasn't caught by SEXYEAR_RE)
    
    Lonely sexagenary binomes without fillers should remain as gz (day), not be
    promoted to sexYear (year).

    :param xml_root: et.Element, root of XML tree to process
    :return: et.Element, modified XML root
    """
    for d in xml_root.xpath(".//date[gz]"):
        prev = d.getprevious()
        has_year_marker = False
        filler_text = ""
        filler_position = "before"  # "before" for 歲次/歲在, "after" for 年/歲 suffix

        # Check for prefix markers (歲次 or 歲在)
        if prev is None:
            s = d.getparent().text or ""
            loc = ("parent", d.getparent())
        else:
            s = prev.tail or ""
            loc = ("tail", prev)

        m = SEX_YEAR_PREFIX_RE.search(s)
        if m:
            has_year_marker = True
            filler_text = m.group(1)
            filler_position = "before"
            # Remove prefix text
            new_s = s[:m.start()]
            if loc[0] == "parent":
                loc[1].text = new_s
            else:
                loc[1].tail = new_s

        # Check for suffix markers (年 or 歲) if no prefix found
        if not has_year_marker:
            tail = d.tail or ""
            # Check if tail starts with 年 or 歲 (possibly with punctuation)
            suffix_match = re.match(r"^([，,\s]*)([年歲])", tail)
            if suffix_match:
                has_year_marker = True
                filler_text = suffix_match.group(2)
                filler_position = "after"
                # Remove the filler from tail
                d.tail = suffix_match.group(1) + tail[suffix_match.end():]

        if not has_year_marker:
            continue

        gz_text = d.findtext("gz")

        # Find the gz element and replace it with sexYear
        gz_elem = d.find("gz")
        if gz_elem is not None:
            # Create sexYear element to replace gz
            sy = et.Element("sexYear")
            sy.text = gz_text

            # Replace gz with sexYear
            d.replace(gz_elem, sy)

            # Add filler element
            if filler_text:
                f = et.Element("filler")
                f.text = filler_text
                if filler_position == "before":
                    d.insert(d.index(sy), f)
                else:  # after
                    d.insert(d.index(sy) + 1, f)

    return xml_root


def promote_nmdgz(xml_root):
    """
    Promote sexagenary day (gz) elements to numbered month day gz (nmdgz) when followed by day elements.

    :param xml_root: et.Element, root of XML tree to process
    :return: et.Element, modified XML root
    """
    for gz_date in list(xml_root.xpath(".//date[gz]")):
        parent = gz_date.getparent()
        gz_text = gz_date.findtext("gz")
        if not gz_text:
            continue

        # ---------- CASE 1 ----------
        # <date><gz>..</gz></date>朔，<date><day>..</day></date>
        tail = gz_date.tail or ""
        if tail.startswith("朔"):
            rest = PUNCT_RE.sub("", tail[1:])
            next_el = gz_date.getnext()

            if next_el is not None and next_el.tag == "date" and next_el.find("day") is not None:
                # Clean tail of gz_date
                gz_date.tail = rest

                # Add nmdgz + lp_filler to day date
                nmdgz = et.SubElement(next_el, "nmdgz")
                nmdgz.text = gz_text
                lp = et.SubElement(next_el, "lp_filler")
                lp.text = "朔"

                # Remove gz_date but preserve its tail
                prev = gz_date.getprevious()
                if prev is None:
                    parent.text = (parent.text or "") + (gz_date.tail or "")
                else:
                    prev.tail = (prev.tail or "") + (gz_date.tail or "")
                parent.remove(gz_date)
                continue

        # ---------- CASE 2 ----------
        # 朔<date><gz>..</gz></date>，<date><day>..</day></date>
        prev = gz_date.getprevious()
        if prev is None:
            s = parent.text or ""
            loc = ("parent", parent)
        else:
            s = prev.tail or ""
            loc = ("tail", prev)

        if s.endswith("朔"):
            next_el = gz_date.getnext()
            if next_el is not None and next_el.tag == "date" and next_el.find("day") is not None:
                # Remove trailing 朔
                new_s = s[:-1]
                if loc[0] == "parent":
                    loc[1].text = new_s
                else:
                    loc[1].tail = new_s

                # Move gz into day date
                nmdgz = et.SubElement(next_el, "nmdgz")
                nmdgz.text = gz_text
                lp = et.SubElement(next_el, "lp_filler")
                lp.text = "朔"

                # Remove gz_date, preserve its tail
                gz_tail = gz_date.tail or ""
                prev2 = gz_date.getprevious()
                if prev2 is None:
                    parent.text = (parent.text or "") + gz_tail
                else:
                    prev2.tail = (prev2.tail or "") + gz_tail
                parent.remove(gz_date)

    return xml_root


def attach_suffixes(xml_root: et.Element) -> et.Element:
    """
    Convert:
      <date><era>太和</era></date>初
    into:
      <date><era>太和</era><suffix>初</suffix></date>

    Same for <ruler> and <dyn>.
    """
    # Snapshot because we mutate tails
    for d in list(xml_root.xpath(".//date")):
        tail = d.tail or ""
        if not tail:
            continue

        # Decide which suffix regex applies based on content
        if d.find("ruler") is not None:
            m = RULER_SUFFIX_RE.match(tail)
        elif d.find("era") is not None:
            m = ERA_SUFFIX_RE.match(tail)
        elif d.find("dyn") is not None:
            m = DYNASTY_SUFFIX_RE.match(tail)
        else:
            continue

        if not m:
            continue

        suf = m.group(1)

        # Add/append suffix element
        s_el = et.SubElement(d, "suffix")
        s_el.text = suf

        # Remove suffix from tail; keep remainder intact
        d.tail = tail[m.end():]

    return xml_root


def tag_date_elements(text, civ=None):
    """
    Tag and clean Chinese string containing date with relevant elements for extraction. Each date element remains
    separated, awaiting "consolidation."
    :param text: str
    :param civ: str ('c', 'j', 'k') or list (['c', 'j', 'k']) to filter by civilization
    :return: str (XML)
    """
    # Test if input is XML, if not, wrap in <root> tags to make it XML
    try:
        xml_root = et.fromstring(text.encode("utf-8"))
    except et.ParseError:
        try:
            xml_root = et.fromstring('<root>' + text + '</root>')
        except et.ParseError:
            # If both parsing attempts fail, create a minimal root element
            xml_root = et.Element("root")
            xml_root.text = text

    # Ensure xml_root is not None
    if xml_root is None:
        xml_root = et.Element("root")
        xml_root.text = text if text else ""
    
    # Defaults
    if civ is None:
        civ = ['c', 'j', 'k']

    # Relational prefixes (unit-based) ###################################################################################
    # Tag early so we don't accidentally tag e.g. "明年" as the Ming dynasty "明".
    # IMPORTANT: bare "其" (unit="") is handled later structurally, not here.
    def make_rel(match):
        dir_ = match.group(1) or ""
        unit = match.group(2) or ""
        comma = (match.group(3) or "")
        rel_text = dir_ + unit + comma

        el = et.Element("rel")
        el.set("dir", dir_)
        el.set("unit", unit)
        el.text = rel_text
        # Note: el.tail is set by replace_in_text_and_tail to preserve text after the match
        return el
    
    def make_xianshi(match):
        """Handle '先是' (previously/before this) - special compound pattern"""
        comma = (match.group(2) or "")
        rel_text = "先是" + comma
        
        el = et.Element("rel")
        el.set("dir", "先")
        el.set("unit", "")  # No unit for "先是"
        el.text = rel_text
        # Note: el.tail is set by replace_in_text_and_tail to preserve text after the match
        
        return el

    # Apply patterns in order: "先是" first (most specific), then "明", then others
    # "先是" must come before REL_RE_OTHER to avoid matching "先" separately
    replace_in_text_and_tail(xml_root, REL_RE_XIANSHI, make_xianshi, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, REL_RE_MING, make_rel, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    replace_in_text_and_tail(xml_root, REL_RE_OTHER, make_rel, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # Retrieve tag tables
    era_tag_df = load_csv('era_table.csv')
    # Filter era_tag_df by cal_stream
    cal_streams = get_cal_streams_from_civ(civ)
    if cal_streams is not None:
        era_tag_df = era_tag_df[era_tag_df['cal_stream'].notna()]
        # Convert cal_stream to float for comparison to avoid int/float mismatch
        era_tag_df = era_tag_df[era_tag_df['cal_stream'].astype(float).isin(cal_streams)]
    dyn_tag_df, ruler_tag_df = load_tag_tables(civ=civ)
    # Reduce to lists
    era_tag_list = era_tag_df['era_name'].unique()
    dyn_tag_list = dyn_tag_df['string'].unique()
    
    # Split ruler tags into regular and era_prefix_only
    if 'era_prefix_only' in ruler_tag_df.columns:
        era_prefix_ruler_tags = ruler_tag_df[ruler_tag_df['era_prefix_only'] == True]['string'].unique()
        regular_ruler_tags = ruler_tag_df[ruler_tag_df['era_prefix_only'] != True]['string'].unique()
    else:
        era_prefix_ruler_tags = []
        regular_ruler_tags = ruler_tag_df['string'].unique()
    ruler_tag_list = list(regular_ruler_tags) + list(era_prefix_ruler_tags)  # Keep for backward compatibility
    # Normal dates #####################################################################################################
    # Tag 改元 first so later passes don't mis-tag the 元 inside it (e.g. as a dynasty).
    def make_gy(match):
        el = et.Element("gy")
        el.text = match.group(1)
        return el
    replace_in_text_and_tail(xml_root, GY_RE, make_gy, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # Year, month, day, gz, season, lp
    xml_root = tag_basic_tokens(xml_root)
    # Lunar phases
    replace_in_text_and_tail(xml_root, LP_RE, make_simple_date("lp"), skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
    # NM date
    xml_root = promote_nmdgz(xml_root)
    # Era names ########################################################################################################
    # Reduce list
    era_tag_list = [s for s in era_tag_list if isinstance(s, str) and s]
    if era_tag_list:
        era_tag_list.sort(key=len, reverse=True)
        era_pattern = re.compile("(" + "|".join(map(re.escape, era_tag_list)) + ")")

        def make_era(match):
            d = et.Element("date")
            e = et.SubElement(d, "era")
            e.text = match.group(1)
            return d

        # First pass: Tag eras immediately before <date> elements
        def tag_eras_before_dates(xml_root, pattern, make_element):
            """Tag era names that occur immediately before <date> elements."""
            changed = True
            max_passes = 10
            for _ in range(max_passes):
                if not changed:
                    break
                changed = False
                
                # Collect only top-level date elements (filter nested ones during collection)
                date_elements = []
                for date_el in xml_root.iter("date"):
                    # Skip if nested inside another date element (check only direct parent)
                    parent = date_el.getparent()
                    if parent is not None and parent.tag == "date":
                        continue
                    date_elements.append(date_el)
                
                for date_el in date_elements:
                    parent = date_el.getparent()
                    if parent is None:
                        continue
                    
                    idx = parent.index(date_el)
                    text_to_check = None
                    is_tail = False
                    target_element = None
                    
                    # Check the tail of the previous sibling (if exists)
                    if idx > 0:
                        prev_sibling = parent[idx - 1]
                        if prev_sibling.tail:
                            text_to_check = prev_sibling.tail
                            is_tail = True
                            target_element = prev_sibling
                    # Otherwise check parent's text before this date element
                    elif parent.text:
                        text_to_check = parent.text
                        is_tail = False
                        target_element = parent
                    
                    if text_to_check:
                        # Use regex to find all matches ending at the end of the text
                        # This is more efficient than iterating through era list
                        matches_at_end = [m for m in pattern.finditer(text_to_check) if m.end() == len(text_to_check)]
                        if matches_at_end:
                            # Take the longest match (first in sorted list is longest)
                            match = matches_at_end[0]
                            new_el = make_element(match)
                            
                            if is_tail:
                                target_element.tail = text_to_check[:match.start()]
                                parent.insert(idx, new_el)
                            else:
                                target_element.text = text_to_check[:match.start()]
                                parent.insert(0, new_el)
                            changed = True
                            # Break to restart iteration with updated structure
                            break
        
        tag_eras_before_dates(xml_root, era_pattern, make_era)
        
        # Second pass: Tag eras anywhere else
        replace_in_text_and_tail(xml_root, era_pattern, make_era, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # Ruler Names ######################################################################################################
    # First pass: Tag era_prefix_only ruler tags immediately before era elements
    if len(era_prefix_ruler_tags) > 0:
        era_prefix_ruler_list = [s for s in era_prefix_ruler_tags if isinstance(s, str) and s]
        if era_prefix_ruler_list:
            era_prefix_ruler_list.sort(key=len, reverse=True)
            era_prefix_ruler_pattern = re.compile("(" + "|".join(map(re.escape, era_prefix_ruler_list)) + ")")

            def make_ruler(match):
                d = et.Element("date")
                e = et.SubElement(d, "ruler")
                e.text = match.group(1)
                return d

            def tag_era_prefix_rulers_before_eras(xml_root, pattern, make_element):
                """Tag era_prefix_only ruler names that occur immediately before <date> elements containing eras."""
                changed = True
                max_passes = 10
                for _ in range(max_passes):
                    if not changed:
                        break
                    changed = False
                    
                    # Find all date elements that contain era elements
                    date_elements_with_era = []
                    for date_el in xml_root.iter("date"):
                        # Check if this date element contains an era
                        if date_el.find("era") is not None:
                            # Skip if nested inside another date element
                            parent = date_el.getparent()
                            if parent is not None and parent.tag == "date":
                                continue
                            date_elements_with_era.append(date_el)
                    
                    for date_el in date_elements_with_era:
                        parent = date_el.getparent()
                        if parent is None:
                            continue
                        
                        idx = parent.index(date_el)
                        text_to_check = None
                        is_tail = False
                        target_element = None
                        
                        # Check the tail of the previous sibling (if exists)
                        if idx > 0:
                            prev_sibling = parent[idx - 1]
                            if prev_sibling.tail:
                                text_to_check = prev_sibling.tail
                                is_tail = True
                                target_element = prev_sibling
                        # Otherwise check parent's text before this date element
                        elif parent.text:
                            text_to_check = parent.text
                            is_tail = False
                            target_element = parent
                        
                        if text_to_check:
                            # Use regex to find all matches ending at the end of the text
                            matches_at_end = [m for m in pattern.finditer(text_to_check) if m.end() == len(text_to_check)]
                            if matches_at_end:
                                # Take the longest match
                                match = matches_at_end[0]
                                new_el = make_element(match)
                                
                                if is_tail:
                                    target_element.tail = text_to_check[:match.start()]
                                    parent.insert(idx, new_el)
                                else:
                                    target_element.text = text_to_check[:match.start()]
                                    parent.insert(0, new_el)
                                changed = True
                                # Break to restart iteration with updated structure
                                break
            
            tag_era_prefix_rulers_before_eras(xml_root, era_prefix_ruler_pattern, make_ruler)
    
    # Second pass: Tag regular ruler tags anywhere (exclude era_prefix_only tags)
    regular_ruler_list = [s for s in regular_ruler_tags if isinstance(s, str) and s]
    if regular_ruler_list:
        regular_ruler_list.sort(key=len, reverse=True)
        ruler_pattern = re.compile("(" + "|".join(map(re.escape, regular_ruler_list)) + ")")

        def make_ruler(match):
            d = et.Element("date")
            e = et.SubElement(d, "ruler")
            e.text = match.group(1)
            return d

        replace_in_text_and_tail(xml_root, ruler_pattern, make_ruler, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)
        
    # Dynasty Names ####################################################################################################
    # Reduce list
    dyn_tag_list = [s for s in dyn_tag_list if isinstance(s, str) and s]
    if dyn_tag_list:
        dyn_tag_list.sort(key=len, reverse=True)
        dyn_pattern = re.compile("(" + "|".join(map(re.escape, dyn_tag_list)) + ")")

        def make_dyn(match):
            d = et.Element("date")
            e = et.SubElement(d, "dyn")
            e.text = match.group(1)
            return d

        replace_in_text_and_tail(xml_root, dyn_pattern, make_dyn, skip_text_tags=SKIP_TEXT_ONLY, skip_all_tags=SKIP_ALL)

    # Sexagenary year promotion (must happen after era/ruler/dynasty tagging)
    xml_root = promote_gz_to_sexyear(xml_root)

    # Suffixes #########################################################################################################
    xml_root = attach_suffixes(xml_root)

    # Bare "其" without unit (structural tagging) #######################################################################
    #
    # Rule: tag standalone "其" only when it is directly adjacent to a following <date>
    # (no punctuation between), and that immediate <date> begins with year/sexYear or lower.
    #
    # This prevents tagging discourse "其，" and prevents gluing "其" onto dynasty/ruler/era-only dates.
    YEAR_OR_LOWER_TAGS = {"year", "sexYear", "month", "day", "gz", "nmdgz", "int", "lp", "season", "lp_filler", "filler"}

    def _date_child_tags(d: et._Element) -> set[str]:
        return {c.tag for c in d if isinstance(c.tag, str)}

    def _maybe_insert_bare_qi_before_next_date(parent: et._Element, before_node: et._Element | None) -> None:
        """
        If the text slot right before the next sibling <date> ends with bare '其' (optionally followed by whitespace),
        replace that '其' with a <rel dir="其" unit="">其</rel> element.
        """
        # Determine the string slot we are checking and the next sibling <date>
        if before_node is None:
            s = parent.text or ""
            next_el = parent[0] if len(parent) > 0 else None
            setter = ("text", parent)
        else:
            s = before_node.tail or ""
            next_el = before_node.getnext()
            setter = ("tail", before_node)

        if not s or next_el is None or next_el.tag != "date":
            return

        # Must end with bare '其' (allow trailing whitespace only)
        m = re.search(r"其(\s*)$", s)
        if not m:
            return

        # The immediate next <date> must begin with year/sexYear or lower
        next_tags = _date_child_tags(next_el)
        if not (next_tags & YEAR_OR_LOWER_TAGS):
            return

        ws = m.group(1) or ""
        prefix = s[:m.start()]

        rel_el = et.Element("rel")
        rel_el.set("dir", "其")
        rel_el.set("unit", "")
        rel_el.text = "其"
        rel_el.tail = ws

        if setter[0] == "text":
            parent.text = prefix
            parent.insert(0, rel_el)
        else:
            before_node.tail = prefix
            # insert after before_node
            idx = parent.index(before_node)
            parent.insert(idx + 1, rel_el)

    for parent in list(xml_root.iter()):
        # Only consider parents that actually have a following element to attach to
        if len(parent) > 0:
            _maybe_insert_bare_qi_before_next_date(parent, None)
            for child in list(parent):
                _maybe_insert_bare_qi_before_next_date(parent, child)

    # Attach/wrap standalone <rel> #########################################################################
    #
    # Rules:
    # - If <rel> immediately precedes a <date>, move it into that <date> as first child ONLY if allowed by rules below.
    # - If <rel> is standalone, keep only those with non-empty unit by wrapping as <date><rel .../></date>.
    # - Otherwise drop standalone rel (unit="").
    #
    # Attachment rules (user rules):
    # - "其" and "先是" with unit="" can only attach if the immediate next <date> begins with year/sexYear or lower.
    # - "是歲" / "其歲" / "今歲" (and "是年"/"其年"/"今年") may precede anything, but can only attach if the *date cluster* has year/sexYear or lower.
    # - "是月" / "其月" / "今月" can only attach if the *date cluster* has month or lower.
    # - All other rel markers require a unit; they can only attach if the immediate next <date> begins with month or lower.
    MONTH_OR_LOWER_TAGS = {"month", "day", "gz", "nmdgz", "int", "lp", "season", "lp_filler", "filler"}
    JOINER_TAIL_RE = re.compile(r"^[，,\s]*$")

    def _collect_date_cluster(start: et._Element) -> list[et._Element]:
        cluster = [start]
        cur = start
        while True:
            tail = cur.tail or ""
            if not JOINER_TAIL_RE.match(tail):
                break
            nxt = cur.getnext()
            if nxt is None or nxt.tag != "date":
                break
            cluster.append(nxt)
            cur = nxt
        return cluster

    def _cluster_child_tags(cluster: list[et._Element]) -> set[str]:
        out: set[str] = set()
        for d in cluster:
            out |= _date_child_tags(d)
        return out

    for rel in list(xml_root.xpath(".//rel")):
        # Skip rel already inside a date
        if rel.xpath("boolean(ancestor::date)"):
            continue

        parent = rel.getparent()
        if parent is None:
            continue

        next_el = rel.getnext()
        rel_tail = rel.tail or ""

        # Case A: immediately before a <date>
        if next_el is not None and next_el.tag == "date" and rel_tail.strip() == "":
            dir_ = (rel.get("dir") or "").strip()
            unit = (rel.get("unit") or "").strip()

            next_tags = _date_child_tags(next_el)
            cluster = _collect_date_cluster(next_el)
            cluster_tags = _cluster_child_tags(cluster)

            attach_ok = False

            # Bare 其 or 先是 (unit="") is only allowed if the immediate next date begins with year/sexYear or lower.
            if unit == "" and dir_ in ("其", "先"):
                attach_ok = bool(next_tags & YEAR_OR_LOWER_TAGS)

            # 是歲 / 其歲 / 今歲 (and 是年/其年/今年) may precede dyn/ruler/era, but only attach if the cluster has year/sexYear or lower.
            elif dir_ in {"是", "其", "今"} and unit in {"歲", "年"}:
                attach_ok = bool(cluster_tags & YEAR_OR_LOWER_TAGS)

            # Optional: 是月 / 其月 / 今月 attach only if the cluster has month or lower.
            elif dir_ in {"是", "其", "今"} and unit == "月":
                attach_ok = bool(cluster_tags & MONTH_OR_LOWER_TAGS)

            # All other rel markers require a unit and only attach if what follows begins with month or lower.
            else:
                if unit != "":
                    attach_ok = bool(next_tags & MONTH_OR_LOWER_TAGS)

            if attach_ok:
                idx = parent.index(rel)
                parent.remove(rel)
                rel.tail = None
                next_el.insert(0, rel)
                continue

        # Case B: standalone rel with unit -> wrap into its own <date>
        unit = rel.get("unit") or ""
        if unit.strip() != "":
            idx = parent.index(rel)
            # Preserve any tail outside the new <date>
            tail_to_preserve = rel.tail
            rel.tail = None

            # Create <date> wrapper and move rel inside it
            d = et.Element("date")
            parent.insert(idx, d)
            parent.remove(rel)
            d.append(rel)

            # Reattach preserved tail to the wrapper <date> (so surrounding text stays in document)
            if tail_to_preserve:
                d.tail = (d.tail or "") + tail_to_preserve
            continue

        # Case C: standalone rel with empty unit -> drop it, preserve tail in parent
        prev = rel.getprevious()
        tail = rel.tail or ""
        if prev is None:
            parent.text = (parent.text or "") + tail
        else:
            prev.tail = (prev.tail or "") + tail
        parent.remove(rel)

    # Clean nested tags ################################################################################################
    # Remove lone tags
    for node in xml_root.xpath('.//date'):
        s = node.xpath('string()')
        bad = ['一年', '一日']
        if s in bad:
            node.tag = 'to_remove'
    # Strip tags
    et.strip_tags(xml_root, 'to_remove')
    # Return to string
    text = et.tostring(xml_root, encoding='utf8').decode('utf8')
    
    return text


def consolidate_date(text):
    """
    Join separated date elements in the XML according to typical date order (year after era, month after year, etc.)
    :param text: str (XML)
    :return: str (XML)
    """
    # Remove spaces
    bu = text
    xml_root = et.ElementTree(et.fromstring(text)).getroot()
    xml_root = strip_ws_in_text_nodes(xml_root)
    text = et.tostring(xml_root, encoding='utf8').decode('utf8')
    ls = [
        ('dyn', 'ruler'),
        ('ruler', 'year'), ('ruler', 'era'),
        ('era', 'year'),
        ('era', 'sexYear'),
        ('era', 'filler'),
        ('ruler', 'filler'),
        ('dyn', 'filler'),
        ('year', 'season'),
        ('year', 'filler'),
        ('year', 'sexYear'),
        ('sexYear', 'season'),
        ('sexYear', 'int'),
        ('sexYear', 'month'),
        ('year', 'int'),
        ('year', 'month'),
        ('season', 'int'),
        ('season', 'month'),
        ('int', 'month'),
        ('month', 'gz'),
        ('month', 'lp'),
        ('month', 'day'),
        ('month', 'nmdgz'),
        ('gz', 'lp'),
        ('nmdgz', 'day'),
        ('day', 'gz'),
        ('month', 'lp_filler'),
        ('lp_filler', 'day'),
        ('gz', 'filler'),
        ('dyn', 'era')
    ]
    for tup in ls:
        text = re.sub(rf'</{tup[0]}></date>，*<date><{tup[1]}', f'</{tup[0]}><{tup[1]}', text)
        if 'metadata' in text:
            text = clean_attributes(text)
    # Parse to XML and return as string
    try:
        et.ElementTree(et.fromstring(text)).getroot()
        return text
    except et.ParseError:
        return "<root/>"


def clean_nested_tags(text):
    """
    Clean nested and invalid date tags from XML string.

    :param text: str, XML string with date tags
    :return: str, cleaned XML string
    """
    xml_root = et.ElementTree(et.fromstring(text)).getroot()
    # Clean
    for node in xml_root.xpath('.//date//date'):
        node.tag = 'to_remove'
    for tag in ['dyn', 'ruler', 'year', 'month', 'season', 'day', 'gz', 'lp', 'sexYear', 'nmdgz', 'lp_to_remove']:
        for node in xml_root.findall(f'.//{tag}//*'):
            node.tag = 'to_remove'
    for node in xml_root.findall('.//date'):
        heads = node.xpath('.//ancestor::head')
        if len(heads) == 0:
            elements = [sn.tag for sn in node.findall('./*')]
            # Clean dynasty only
            if elements == ['dyn'] or elements == ['season'] or elements == ['era'] or elements == ['ruler']:
                for sn in node.findall('.//*'):
                    sn.tag = 'to_remove'
                node.tag = 'to_remove'
    # Strip tags
    et.strip_tags(xml_root, 'to_remove')
    et.strip_tags(xml_root, 'lp_to_remove')
    # Return to string
    text = et.tostring(xml_root, encoding='utf8').decode('utf8')
    return text


def index_date_nodes(xml_root) -> et._Element:
    """
    Index date nodes in XML element.
    """
    # Handle namespaces
    ns = {}
    if xml_root.tag.startswith('{'):
        ns_uri = xml_root.tag.split('}')[0][1:]
        ns = {'tei': ns_uri}

    index = 0
    date_xpath = './/tei:date' if ns else './/date'
    for node in xml_root.xpath(date_xpath, namespaces=ns):
        node.set('index', str(index))
        index += 1

    return xml_root



import re
import pandas as pd
from .converters import (
    gz_year, ganshu, numcon, iso_to_jdn, jdn_to_iso, jdn_to_gz
)
from .config import (
    phrase_dic_en, get_phrase_dic
)
from .loaders import (
    prepare_tables, load_csv
)
from .utils import guess_variant


def generate_report_from_dataframe(output_df, phrase_dic=phrase_dic_en, jd_out=False, tpq=None, taq=None):
    """
    Generate human-readable report from processed dataframe.

    :param output_df: DataFrame with processed date results (includes error_str, date_string, etc.)
    :param phrase_dic: Dictionary with UI phrases
    :param jd_out: Whether to output Julian Day numbers
    :param tpq: Terminus post quem (earliest date filter)
    :param taq: Terminus ante quem (latest date filter)
    :return: Formatted report string
    """
    if output_df.empty:
        return f'{phrase_dic["ui"]}: {phrase_dic["unknown-date"]}\n{phrase_dic["matches"]}:\n{phrase_dic["no-matches"]}'

    # Add missing columns with NaN values (so they can be used in boolean operations)
    cols = ["dyn_name", "ruler_name", "era_name", "year", "sex_year", "month", "intercalary", "day", "gz", "lp", "nmd_gz"]
    for col in cols:
        if col not in output_df.columns:
            output_df[col] = pd.NA
    
    # Check if any rows have resolved historical entities (dyn_id, ruler_id, or era_id)
    has_resolved_entities = (
        ('dyn_id' in output_df.columns and output_df['dyn_id'].notna().any()) or
        ('ruler_id' in output_df.columns and output_df['ruler_id'].notna().any()) or
        ('era_id' in output_df.columns and output_df['era_id'].notna().any())
    )

    if not has_resolved_entities:
        # Format as a proper report entry
        if not output_df.empty and 'date_string' in output_df.columns:
            date_string = output_df['date_string'].iloc[0]
        else:
            date_string = "unknown date"
        return f'{phrase_dic["ui"]}: {date_string}\n{phrase_dic["matches"]}:\n{phrase_dic["insuff-data"]}'

    # Check for too many candidates - filter by tpq/taq if provided
    if len(output_df) > 15:
        # Filter by tpq/taq if provided
        if tpq is not None or taq is not None:
            # Use ind_year if available, otherwise use era_start_year or era_end_year
            filter_mask = pd.Series([True] * len(output_df), index=output_df.index)
            
            if 'ind_year' in output_df.columns:
                if tpq is not None:
                    filter_mask &= output_df['ind_year'].notna() & (output_df['ind_year'] >= tpq)
                if taq is not None:
                    filter_mask &= output_df['ind_year'].notna() & (output_df['ind_year'] <= taq)
            elif 'era_start_year' in output_df.columns:
                if tpq is not None:
                    filter_mask &= output_df['era_start_year'].notna() & (output_df['era_start_year'] >= tpq)
                if taq is not None:
                    filter_mask &= output_df['era_start_year'].notna() & (output_df['era_start_year'] <= taq)
            elif 'era_end_year' in output_df.columns:
                if tpq is not None:
                    filter_mask &= output_df['era_end_year'].notna() & (output_df['era_end_year'] >= tpq)
                if taq is not None:
                    filter_mask &= output_df['era_end_year'].notna() & (output_df['era_end_year'] <= taq)
            
            output_df = output_df[filter_mask].copy()
        
        # If still too many after filtering, return error message
        if len(output_df) > 15:
            date_string = output_df['date_string'].iloc[0] if not output_df.empty and 'date_string' in output_df.columns else "unknown date"
            return f'{phrase_dic["ui"]}: {date_string}\n{phrase_dic["matches"]}:\n{len(output_df)} {phrase_dic["too-many-cand"]}'

    # Prepare dataframe for vectorized formatting
    df = output_df.copy()

    # If this date has a relative year marker, append a simple warning to error_str
    # (we keep using existing rel_* columns; no new schema required).
    if all(c in df.columns for c in ["rel_dir", "rel_unit"]):
        rel_dir = df["rel_dir"].astype("string").str.strip()
        rel_unit = df["rel_unit"].astype("string").str.strip()
        relative_dirs = {"明", "來", "次", "去", "昨", "前", "後"}
        is_relative_year = rel_unit.isin(["年", "歲"]) & rel_dir.isin(list(relative_dirs))
        if is_relative_year.any():
            if "error_str" not in df.columns:
                df["error_str"] = ""
            # Avoid duplicating if upstream already added the warning.
            existing = df.loc[is_relative_year, "error_str"].fillna("")
            needs = ~existing.str.contains("relative date", regex=False)
            df.loc[is_relative_year & needs, "error_str"] = existing[needs] + "relative date"

    # Ensure strings for concatenation (handle missing columns)
    for col in ["dyn_name", "ruler_name", "era_name"]:
        if col in df.columns:
            df[col] = df[col].fillna("")
        else:
            df[col] = ""

    # Format year strings
    df["year_str"] = ""
    mask = df["year"].notna()
    df.loc[mask & (df["year"] == 1), "year_str"] = "元年"
    df.loc[mask & (df["year"] != 1), "year_str"] = df.loc[mask & (df["year"] != 1) & df["year"].notna(), "year"].astype(int).map(lambda x: str(numcon(x)) + "年")
    
    # Format month strings
    df["month_str"] = ""
    m = df["month"].notna()
    df.loc[m & (df["month"] == 1), "month_str"] = "正月"
    df.loc[m & (df["month"] == 13), "month_str"] = "臘月"
    df.loc[m & (df["month"] == 14), "month_str"] = "一月"
    df.loc[m & ~df["month"].isin([1, 13, 14]), "month_str"] = df.loc[m & ~df["month"].isin([1, 13, 14]), "month"].astype(int).map(lambda x: str(numcon(x)) + "月")

    # Intercalary marker
    df["int_str"] = ""
    df.loc[df.get("intercalary", pd.Series(0, index=df.index, dtype=int)) == 1, "int_str"] = "閏"

    # Day strings
    df["day_str"] = ""
    d = (df["day"].notna()) & (df["lp"] != 0)
    df.loc[d, "day_str"] = df.loc[d & df["day"].notna(), "day"].astype(int).map(lambda x: str(numcon(x)) + "日")

    # Sexagenary day
    df["gz_str"] = ""
    gz_mask = df["gz"].notna()
    df.loc[gz_mask, "gz_str"] = df.loc[gz_mask & df["gz"].notna(), "gz"].astype(int).map(ganshu)

    # Lunar phase
    df["lp_str"] = ""
    lp_mask = df["lp"].notna()
    lp_d = {0: '朔', -1: '晦'}
    df.loc[lp_mask, "lp_str"] = df.loc[lp_mask & df["lp"].notna(), "lp"].astype(int).map(lambda x: lp_d.get(x, ''))

    # Date range strings
    df["range_str"] = ""

    # Era year spans for dynasty/ruler/era only dates (no year specified)
    try:
        if (not df.empty and "era_start_year" in df.columns and "era_end_year" in df.columns and
            "year" in df.columns and "month" in df.columns and "day" in df.columns and
            "gz" in df.columns and "lp" in df.columns):
            era_only_mask = (
                df["era_start_year"].notna() & df["era_end_year"].notna() &
                df["year"].isna() & df["month"].isna() & df["day"].isna() & df["gz"].isna() & df["lp"].isna()
            )
            df.loc[era_only_mask, "range_str"] = (
                "年間（" + df.loc[era_only_mask & df["era_start_year"].notna() & df["era_end_year"].notna(), "era_start_year"].astype(int).astype(str) +
                "–" + df.loc[era_only_mask & df["era_start_year"].notna() & df["era_end_year"].notna(), "era_end_year"].astype(int).astype(str) + "）"
            )
        # else: no era processing needed
    except Exception:
        # If anything goes wrong with era processing, skip it
        pass
    
    # Lunar month ranges (for month-only dates)
    # Only create lunar range strings if the necessary lunar columns exist
    if all(col in df.columns for col in ["nmd_gz", "ISO_Date_Start", "start_gz", "ISO_Date_End", "end_gz"]):
        lunar_range_mask = (
            df["month"].notna() & df["day"].isna() & df["gz"].isna() & df["lp"].isna() &
            df["nmd_gz"].notna() & df["ISO_Date_Start"].notna()
        )
        if jd_out:
            df.loc[lunar_range_mask, "range_str"] = (
                "（JD " + df.loc[lunar_range_mask, "nmd_jdn"].astype(str) +
                " ~ " + df.loc[lunar_range_mask, "hui_jdn"].astype(str) + "）"
            )
        else:
            df.loc[lunar_range_mask, "range_str"] = (
                "（" + df.loc[lunar_range_mask, "start_gz"] +
                df.loc[lunar_range_mask, "ISO_Date_Start"] +
                " ~ " + df.loc[lunar_range_mask, "end_gz"] +
                df.loc[lunar_range_mask, "ISO_Date_End"] + "）"
            )

    # Final date strings
    df["jdn_str"] = ""

    # For dates with specific JDN (calculated from day/gz/lp)
    if "jdn" in df.columns:
        jdn_mask = df["jdn"].notna()
        if jd_out:
            df.loc[jdn_mask, "jdn_str"] = "（JD " + df.loc[jdn_mask, "jdn"].astype(str) + "）"
        elif "ISO_Date" in df.columns:
            iso_mask = jdn_mask & df["ISO_Date"].notna()
            df.loc[iso_mask, "jdn_str"] = "（" + df.loc[iso_mask, "ISO_Date"] + "）"

    # For dates with years but no specific JDN, show western year
    if "ind_year" in df.columns:
        year_only_mask = (
            df["ind_year"].notna() &
            (df["jdn"].isna() if "jdn" in df.columns else True) &
            df["year"].notna() & df["month"].isna() & df["day"].isna() & df["gz"].isna() & df["lp"].isna()
        )
        df.loc[year_only_mask, "jdn_str"] = "（" + df.loc[year_only_mask & df["ind_year"].notna(), "ind_year"].astype(int).astype(str) + "）"

    # Sexagenary year (like in jy_to_ccs)
    df["sex_year_str"] = ""
    if "ind_year" in df.columns:
        sex_year_mask = df["ind_year"].notna()
        df.loc[sex_year_mask, "sex_year_str"] = df.loc[sex_year_mask & df["ind_year"].notna(), "ind_year"].astype(int).map(lambda y: f"（歲在{ganshu(gz_year(y))}）")

    # Combine all components into report_line
    df["report_line"] = (
        df["dyn_name"] + df["ruler_name"] + df["era_name"] +
        df["year_str"] + df["sex_year_str"] + df["int_str"] + df["month_str"] +
        df["day_str"] + df["gz_str"] + df["lp_str"] +
        df["range_str"] + df["jdn_str"]
    )

    # Group by date_index and combine lines
    # Deduplicate report_line values to avoid showing the same line multiple times
    lines_by_date = (
        df.groupby("date_index")["report_line"]
        .agg(lambda s: "\n".join([x for x in s.unique() if x]))
    )

    # Generate final report with headers and errors
    # Group by date_index to get metadata for each date
    metadata_by_date = df.groupby("date_index").agg({
        "date_string": "first",  # All rows for same date_index have same date_string
        "error_str": lambda x: next((s for s in x if s), "")  # Get first non-empty error_str
    })

    report_blocks = []
    for idx, meta in metadata_by_date.iterrows():
        header = f'{phrase_dic["ui"]}: {meta["date_string"]}\n{phrase_dic["matches"]}:\n'
        body = lines_by_date.get(idx, "")
        err = meta["error_str"] or ""
        block = header + (body + "\n" if body else "") + err
        report_blocks.append(block.rstrip())

    return "\n\n".join(report_blocks)


def jdn_to_ccs(x, by_era=True, proleptic_gregorian=False, gregorian_start=None, lang='en', civ=None):
    """
    Convert Julian Day Number to Chinese calendar string.
    :param x: float (Julian Day Number) or str (ISO date string Y-M-D)
    :param by_era: bool (filter from era JDN vs index year)
    :param proleptic_gregorian: bool
    :param gregorian_start: list
    :param lang: str, language ('en', 'fr', 'zh', 'ja', 'de'). Defaults to 'en' if not specified or invalid.
    :param civ: str ('c', 'j', 'k') or list (['c', 'j', 'k']) to filter by civilization
    :return output_string: str
    """
    # Defaults
    if gregorian_start is None:
        gregorian_start = [1582, 10, 15]
    if civ is None:
        civ = ['c', 'j', 'k']
    if lang is None:
        lang = 'en'
    phrase_dic = get_phrase_dic(lang)
    if isinstance(x, str):
        iso = x
        jdn = iso_to_jdn(x, proleptic_gregorian, gregorian_start)
    else:
        jdn = x
        iso = jdn_to_iso(jdn, proleptic_gregorian, gregorian_start)
    output_string = f'{phrase_dic.get("ui")}: {iso} (JD {jdn})\n{phrase_dic.get("matches")}:\n'
    # Load CSV tables
    era_df, dyn_df, ruler_df, lunar_table, dyn_tag_df, ruler_tag_df, ruler_can_names = prepare_tables(civ=civ)
    # Filter ruler_tag_df by filtered rulers
    if not ruler_df.empty:
        valid_person_ids = ruler_df['person_id'].unique()
        ruler_tag_df = ruler_tag_df[ruler_tag_df['person_id'].isin(valid_person_ids)]
    # Filter lunar table by JDN
    lunar_table = lunar_table[(lunar_table['nmd_jdn'] <= jdn) & (lunar_table['hui_jdn'] + 1 > jdn)]
    #
    if by_era:
        # Filter era dataframe by JDN
        df = era_df[(era_df['era_start_jdn'] <= jdn) & (era_df['era_end_jdn'] > jdn)].drop_duplicates(subset=['era_id'])
        df = df[['dyn_id', 'cal_stream', 'era_id', 'ruler_id', 'era_name', 'era_start_year']].rename(columns={'ruler_id': 'person_id'})
        # Get ruler names
        df = df.merge(ruler_can_names, how='left', on='person_id')
        df = df.rename(columns={'person_id': 'ruler_id', 'string': 'ruler_name'})
        # Get dynasty names
        df = df.merge(dyn_df[['dyn_id', 'dyn_name']], how='left', on='dyn_id')
        # Merge with lunar table
        lunar_table = df.merge(lunar_table, how='left', on='cal_stream')
        # Add ruler start year, just to be safe
        temp = ruler_df[['person_id', 'emp_start_year']]
        temp = temp.rename(columns={'person_id': 'ruler_id'})
        lunar_table = lunar_table.merge(temp, how='left', on='ruler_id')
    else:
        # Merge dynasties
        lunar_table = lunar_table.merge(dyn_df, how='left', on='cal_stream')
        # Filter by index year
        lunar_table = lunar_table[lunar_table['dyn_start_year'] <= lunar_table['ind_year']]
        lunar_table = lunar_table[lunar_table['dyn_end_year'] > lunar_table['ind_year']]
        del lunar_table['dyn_start_year'], lunar_table['dyn_end_year']
        # Merge rulers
        del ruler_df['cal_stream'], ruler_df['max_year']
        lunar_table = lunar_table.merge(ruler_df, how='left', on='dyn_id')
        # Merge ruler tags
        lunar_table = lunar_table.merge(ruler_tag_df, how='left', on='person_id')
        lunar_table = lunar_table.rename(columns={'person_id': 'ruler_id', 'string': 'ruler_name'})
        # Filter by index year
        lunar_table = lunar_table[lunar_table['emp_start_year'] <= lunar_table['ind_year']]
        lunar_table = lunar_table[lunar_table['emp_end_year'] > lunar_table['ind_year']]
        del lunar_table['emp_end_year']
        # Clean eras
        del era_df['max_year']
        era_df = era_df.drop_duplicates(subset=['era_id'])
        # Merge eras
        lunar_table = lunar_table.merge(era_df, how='left', on=['dyn_id', 'cal_stream', 'ruler_id'])
        # Filter by index year
        lunar_table = lunar_table[lunar_table['era_start_year'] <= lunar_table['ind_year']]
        lunar_table = lunar_table[lunar_table['era_end_year'] > lunar_table['ind_year']]
        del lunar_table['era_end_year']
    if not lunar_table.empty:
        lunar_table = lunar_table.sort_values(by=['cal_stream', 'dyn_id'])
        lunar_table = lunar_table.drop_duplicates(subset=['era_id'])
        # Create strings
        for index, row in lunar_table.iterrows():
            # Output dynasty and ruler name
            output_string += f"{row['dyn_name']}{row['ruler_name']}"
            # Find Julian year
            iso_string = jdn_to_iso(jdn, proleptic_gregorian, gregorian_start)
            if iso_string[0] == '-':
                iso_string = iso_string[1:]
                mult = -1
            else:
                mult = 1
            year = int(re.split('-', iso_string)[0]) * mult
            # Convert to era or ruler year
            # Check if era_start_year is valid (not NaN) - works for both int and float
            if pd.notna(row['era_start_year']):
                # We have a valid era, use it (even if era_name is blank)
                if isinstance(row['era_name'], str) and row['era_name'] != '':
                    output_string += f"{row['era_name']}"
                # Find era year
                era_year = year - int(row['era_start_year']) + 1
                era_year = numcon(era_year) + '年'
                if era_year == "一年":
                    era_year = "元年"
                output_string += era_year
            else:
                # No valid era, fall back to ruler start year
                ruler_year = year - row['emp_start_year'] + 1
                ruler_year = numcon(ruler_year) + '年'
                if ruler_year == "一年":
                    ruler_year = "元年"
                output_string += ruler_year
            # Sexegesimal year
            sex_year = ganshu(row['year_gz'])
            output_string += f"（歲在{sex_year}）"
            # Month
            if row['intercalary'] == 1:
                output_string += '閏'
            if row['month'] == 1:
                month = '正月'
            elif row['month'] == 13:
                month = '臘月'
            elif row['month'] == 14:
                month = '一月'
            else:
                month = numcon(row['month']) + '月'
            output_string += month
            # Find day
            if int(jdn - .5) + .5 == row['nmd_jdn']:
                day = '朔'
            elif int(jdn - .5) + .5 == row['hui_jdn']:
                num = numcon(row['hui_jdn'] - row['nmd_jdn'] + 1) + '日'
                day = f"晦（{num}）"
            else:
                day = numcon(int(jdn - row['nmd_jdn']) + 1) + '日'
            output_string += day
            # Sexagenary day
            output_string += jdn_to_gz(jdn)
            # Line break
            output_string += '\n'
        output_string = output_string[:-1]
        # Output
        return output_string
    else:
        return None


def jy_to_ccs(y, lang='en', civ=None):
    """
    Convert Western year to Chinese calendar string.
    :param y: int
    :param lang: str, language ('en', 'fr', 'zh', 'ja', 'de'). Defaults to 'en' if not specified or invalid.
    :param civ: str ('c', 'j', 'k') or list (['c', 'j', 'k']) to filter by civilization
    :return output_string: str
    """
    # Defaults
    if civ is None:
        civ = ['c', 'j', 'k']
    if lang is None:
        lang = 'en'
    phrase_dic = get_phrase_dic(lang)
    
    # Year format strings for different languages
    if y > 0:
        if lang == 'en':
            fill = f"A.D. {int(y)}"
        elif lang == 'fr':
            fill = f"{int(y)} apr. J.-C."
        elif lang == 'de':
            fill = f"{int(y)} n. Chr."
        elif lang == 'zh':
            fill = f"公元{int(y)}年"
        elif lang == 'ja':
            fill = f"西暦{int(y)}年"
        else:
            fill = f"A.D. {int(y)}"  # Default to English
    else:
        if lang == 'en':
            fill = f"{int(abs(y)) + 1} B.C."
        elif lang == 'fr':
            fill = f"{int(abs(y)) + 1} av. J.-C."
        elif lang == 'de':
            fill = f"{int(abs(y)) + 1} v. Chr."
        elif lang == 'zh':
            fill = f"公元前{int(abs(y)) + 1}年"
        elif lang == 'ja':
            fill = f"紀元前{int(abs(y)) + 1}年"
        else:
            fill = f"{int(abs(y)) + 1} B.C."  # Default to English
    output_string = f'{phrase_dic.get("ui")}: {y} ({fill})\n{phrase_dic.get("matches")}:\n'
    # Load CSV tables
    era_df, dyn_df, ruler_df, lunar_table, dyn_tag_df, ruler_tag_df, ruler_can_names = prepare_tables(civ=civ)
    # Filter ruler_tag_df by filtered rulers
    if not ruler_df.empty:
        valid_person_ids = ruler_df['person_id'].unique()
        ruler_tag_df = ruler_tag_df[ruler_tag_df['person_id'].isin(valid_person_ids)]
    ruler_tag_df = ruler_tag_df[['person_id', 'string']]
    # Filter dynasties by year
    df = dyn_df[(dyn_df['dyn_start_year'] <= y) & (dyn_df['dyn_end_year'] >= y)]
    cols = ['dyn_id', 'dyn_name', 'cal_stream']
    df = df[cols]
    # Merge rulers
    del ruler_df['cal_stream']
    df = df.merge(ruler_df, how='left', on=['dyn_id'])
    # Filter by year
    df = df[(df['emp_start_year'] <= y) & (df['emp_end_year'] >= y)]
    # Merge ruler strings
    df = df.merge(ruler_tag_df, how='left', on='person_id')
    df = df.rename(columns={'person_id': 'ruler_id', 'string': 'ruler_name'})
    cols = ['dyn_id', 'dyn_name', 'cal_stream', 'ruler_id', 'emp_start_year', 'ruler_name']
    df = df[cols]
    # Merge era
    era_df = era_df[['era_id', 'ruler_id', 'era_name', 'era_start_year', 'era_end_year']]
    df = df.merge(era_df, how='left', on='ruler_id')
    # Filter by year
    df = df[(df['era_start_year'] <= y) & (df['era_end_year'] >= y)].sort_values(by=['cal_stream', 'dyn_id'])
    # Filter duplicates
    try:
        df['variant_rank'] = df['era_name'].apply(guess_variant)
        df = (
            df.sort_values(by='variant_rank')
            .drop_duplicates(subset=['ruler_id', 'era_id'], keep="first")
            .drop(columns="variant_rank")
        )
    except TypeError:
        df = df.drop_duplicates(subset=['ruler_id', 'era_id'], keep="first")
    if not df.empty:
        # Create strings
        for index, row in df.iterrows():
            # Output dynasty and ruler name
            output_string += f"{row['dyn_name']}{row['ruler_name']}"
            # Convert to era or ruler year
            if isinstance(row['era_name'], str):
                output_string += f"{row['era_name']}"
                # Find era year
                era_year = y - row['era_start_year'] + 1
                era_year = numcon(era_year) + '年'
                if era_year == "一年":
                    era_year = "元年"
                output_string += era_year
            else:
                ruler_year = y - row['emp_start_year'] + 1
                ruler_year = numcon(ruler_year) + '年'
                if ruler_year == "一年":
                    ruler_year = "元年"
                output_string += ruler_year
            # Sexegesimal year
            sex_year = ganshu(gz_year(y))
            output_string += f"（歲在{sex_year}）"
            # Line break
            output_string += '\n'
        output_string = output_string[:-1]
        # Output
        return output_string
    else:
        return None